from __future__ import annotations


from hx_markup.element import NodeText
from markdown import markdown
from ormspace import model as md


from hx_markup import Element
from typing_extensions import Self

from spacestar import detail_url_template_var, htmx_indicator_template_var, htmx_target_template_var, \
    htmx_url_template_var


class SpaceModel(md.Model):
    
    async def setup_instance(self):
        await self.set_instance_dependencies()

    
    def element_detail(self) -> Element:
        container: Element = Element('div', id=self.tablekey)
        container.children.append(Element('h3', children=str(self)))
        for k, v in self.model_fields.items():
            if value:= getattr(self, k):
                container.children.append(f"""
                ##### {v.title or k}
                {markdown(value)}
                """)
        # container.children.append(Element('ul', '.list-group', children=[Element('li','.list-group-item', children=f'{markdown(f"""# {k}""")}') for k, v in dict(self).items() if v]))
        return container
    
    def element_list_group_item(self) -> Element:
        return Element('li', '.list-group-item', NodeText(str(self)))
    
    def element_list_group_item_action(self):
        template = detail_url_template_var.get()
        return Element('li', '.list-group-item',
                       Element('a', '.list-group-item-action',
                               href=template.format(item_name=self.item_name(), item_key=self.key),
                               children=str(self)))

    def element_list_group_item_htmx_action(self):
        template = htmx_url_template_var.get()
        target = htmx_target_template_var.get()
        indicator = htmx_indicator_template_var.get()
        return Element('li', '.list-group-item', Element('a', '.list-group-item-action', htmx={
                'target': target.format(item_name=self.item_name()),
                'get': template.format(item_name=self.item_name(), item_key=self.key),
                'indicator': indicator.format(item_name=self.item_name())
        }, children=str(self)))
    
    @classmethod
    def element_list_group(cls, instances: list[Self]):
        return Element('ul', '.list-group', children=[
                item.element_list_group_item() for item in instances
        ])
    
    
class SpaceSearchModel(SpaceModel, md.SearchModel):
    pass


from __future__ import annotations

from abc import ABC, abstractmethod

from hx_markup import Element, functions
from hx_markup.element import Div, NodeText
from jinja2 import Template
from lxml import etree
from lxml.builder import E
from markupsafe import Markup
from ormspace.alias import QUERIES
from ormspace.model import getmodel
from starlette.requests import Request
from starlette.responses import HTMLResponse

from spacestar.app import SpaceStar



class ResponseEngine(ABC):
    def __init__(self, request: Request, template_path: str = None, source: str = None):
        self.request = request
        self.app: SpaceStar = self.request.app
        self.template_path = template_path
        self.source = source
        
    @property
    def engine(self) -> Template:
        if self.template_path and self.app.templates_directory:
            return self.app.templates.get_template(self.template_path)
        elif self.source:
            return self.app.from_string(self.source)
        return self.app.index
    
    @abstractmethod
    async def template_data(self) -> dict:
        raise NotImplementedError
    
    async def run(self):
        return self.engine.render(request=self.request, **await self.template_data())

        
class ModelResponse(ResponseEngine):
    def __init__(self, request: Request, *args, **kwargs):
        super().__init__(request, *args, **kwargs)
        self.model = getmodel(self.request.path_params.get('item_name'))

    async def update_dependencies(self, lazy=False, queries: QUERIES | None = None):
        await self.model.update_dependencies_context(queries=queries, lazy=lazy)
    
    @property
    def fields(self):
        return self.model.model_fields.values()
    
    @property
    def query(self):
        result = self.model.query_from_request(request=self.request)
        return result
    
    @property
    def path(self):
        return self.request.url.path
    
    @property
    def field_names(self):
        return self.model.model_fields.keys()
    
    async def instances(self, lazy=False):
        return await self.model.sorted_instances_list(query=self.query, lazy=lazy)

    async def run(self):
        if self.template_path:
            return HTMLResponse(self.app.render(self.request, template=self.template_path, **await self.template_data()))
        elif self.source:
            return HTMLResponse(self.app.templates.from_string(self.source).render(request=self.request, **await self.template_data()))
        return self.app.response(self.request, **await self.template_data())

    
    async def template_data(self) -> dict:
        return {'model': self.model}
    

    
class ListResponse(ModelResponse):
    
    async def template_data(self) -> dict:
        return {
                'header': Element('header', Div('.container-fluid', Element('h1', NodeText(self.app.title)))),
                'content': Markup(etree.tounicode(
                        E.div(
                                E.h2(f'lista de {self.model.plural()}'),
                                E.ul(*[E.li(f'{i.key} {i}', {'class': 'list-group-item text-white'}) for i in
                                       await self.instances()],
                                     {'class': 'list-group', 'style': 'overflow-y: auto; max-height: 80%;'})
                        
                        ))
                        
                ),
                'footer': etree.tounicode(E.footer(f'resultados para {functions.write_args(self.query.values())}', id='footer'))
        }
        

class SearchResponse(ModelResponse):
    async def element(self):
        page = E.div(
                E.h4('resultado de pesquisa de {}'.format(self.model).title()),
        )
        return page
    




from contextlib import contextmanager
from contextvars import ContextVar

detail_url_template_var = ContextVar('detail_url_template_var', default='/api/{}/{}')
htmx_url_template_var = ContextVar('htmx_url_template_var', default='/api/{}/{}')
htmx_target_template_var = ContextVar('htmx_target_template_var', default='#htmx-{}-container')
htmx_indicator_template_var = ContextVar('htmx_indicator_template_var', default='#htmx-{}-indicator')

@contextmanager
def detail_url_context(template: str):
    token = detail_url_template_var.set(template)
    try:
        yield
    finally:
        detail_url_template_var.reset(token)
        
@contextmanager
def htmx_url_context(template: str):
    token = htmx_url_template_var.set(template)
    try:
        yield
    finally:
        htmx_url_template_var.reset(token)


@contextmanager
def htmx_target_context(template: str):
    token = htmx_target_template_var.set(template)
    try:
        yield
    finally:
        htmx_target_template_var.reset(token)
        
@contextmanager
def htmx_indicator_context(template: str):
    token = htmx_indicator_template_var.set(template)
    try:
        yield
    finally:
        htmx_indicator_template_var.reset(token)


if __name__ == '__main__':
    with htmx_url_context('/doctor/{}/{}'):
        with htmx_target_context('#htmx-{}'):
            print(htmx_url_template_var.get().format('invoice', 'xdrerer'))
            print(htmx_target_template_var.get().format('invoice'))
    



import os

import hx_markup
import spacestar


def create_templates_directory():
    os.makedirs(os.path.join(os.getcwd(), 'templates'), exist_ok=True)
    with open(os.path.join(os.getcwd(), 'templates', 'index_from_string.html'), 'w') as f:
        f.write('<!doctype html>')
        f.write(str(hx_markup.Element('html', children=[
                hx_markup.Element('head', children=[spacestar.META_CHARSET, spacestar.META_VIEWPORT,spacestar.BOOTSTRAP_LINK, spacestar.HEAD_STYLE_LINK, spacestar.HEAD_SCRIPT]),
                hx_markup.Element('body', 'body', children=[
                        hx_markup.Element('nav', 'nav'),
                        hx_markup.Element('main', 'main'),
                        hx_markup.Element('footer', 'footer'),
                        '<script type="text/javascript" src="/static/js/body.js"></script>'
                ])
        ])))

def create_static_directory():
    os.makedirs(os.path.join(os.getcwd(), 'static/css'), exist_ok=True)
    os.makedirs(os.path.join(os.getcwd(), 'static/js'), exist_ok=True)
    os.makedirs(os.path.join(os.getcwd(), 'static/img'), exist_ok=True)
    with open(os.path.join(os.getcwd(), 'static', 'css/main.css'), 'w') as f:
        f.write('html, body {padding: 0; margin: 0; box-sizing: border-box}')
    with open(os.path.join(os.getcwd(), 'static', 'js/main.js'), 'w') as f:
        f.write('const body = document.getElementById("body")')
        
        
def main():
    create_templates_directory()
    create_static_directory()
    
if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spacestar']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'Markdown>=3.5.2,<4.0.0',
 'hx-markup>=0.2.2,<0.3.0',
 'itsdangerous>=2.1.2,<3.0.0',
 'ormspace>=0.3.0,<0.4.0',
 'python-multipart>=0.0.6,<0.0.7',
 'starlette>=0.32.0,<0.33.0',
 'uvicorn>=0.24.0,<0.25.0']

entry_points = \
{'console_scripts': ['init-static = '
                     'spacestar.initfolders:create_static_directory',
                     'init-templates = '
                     'spacestar.initfolders:create_templates_directory',
                     'initfolders = spacestar.initfolders:main']}

setup_kwargs = {
    'name': 'spacestar',
    'version': '0.2.8',
    'description': 'Framework for building web based apps with the power of Starlette, Pydantic and Deta Space.',
    'long_description': None,
    'author': 'Daniel Arantes',
    'author_email': 'arantesdv@me.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from contextvars import ContextVar
from contextlib import asynccontextmanager
from typing import Any

context_data = ContextVar("context_data", default={})

@asynccontextmanager
async def load_context(data: dict[str, Any]):
    token = context_data.set(data)
    try:
        yield
    finally:
        context_data.reset(token)


from __future__ import annotations

import os.path
from collections import defaultdict
from contextlib import asynccontextmanager
from contextvars import ContextVar
from functools import wraps
from typing import Any, Optional

import jinja2
import uvicorn
from hx_markup import Element
from markupsafe import Markup
from ormspace import functions
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.requests import Request
from starlette.responses import HTMLResponse
from starlette.routing import Mount, Route, Router
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates
from starlette.datastructures import FormData
from ormspace.settings import Settings

from spacestar.model import SpaceModel

response_contextvar = ContextVar("response_contextvar", default={})

def app_context(request: Request) -> dict[str, Any]:
    return {'app': request.app}

@asynccontextmanager
async def response_context(request: Request, **kwargs):
    response_token = response_contextvar.set(kwargs)
    try:
        yield request.app.response(request, **response_contextvar.get())
    finally:
        response_contextvar.reset(response_token)



class SpaceStar(Starlette):
    """SpaceStar class is a mix of Starlette and Uvicorn for faster run of HTTP server configured for Deta Space.
    Parameters:
    module - A string with the module name running the application. Defaults to "main".
    app_name - A string with the SpaceStar instance name. Defaults to "app".
    lang - Language of the application. Defaults to "en".
    title - The title for home page. Defaults to "SpaceStar".
    static_directory - A string indicating the location of static folder, relative to working directory.
    templates_directory - A string indicating the location of jinja2 templates_directory.
    index_template - index template as a string.
    debug - Boolean indicating if debug tracebacks should be returned on errors.
    routes - A list of routes to serve incoming HTTP and WebSocket requests.
    middleware - A list of middleware to run for every request. A starlette application will always automatically include two middleware classes. ServerErrorMiddleware is added as the very outermost middleware, to handle any uncaught errors occurring anywhere in the entire stack. ExceptionMiddleware is added as the very innermost middleware, to deal with handled exception cases occurring in the routing or endpoints.
    exception_handlers - A mapping of either integer status codes, or exception class types onto callables which handle the exceptions. Exception handler callables should be of the form handler(request, exc) -> response and may be either standard functions, or async functions.
    on_startup - A list of callables to run on application startup. Startup handler callables do not take any arguments, and may be either standard functions, or async functions.
    on_shutdown - A list of callables to run on application shutdown. Shutdown handler callables do not take any arguments, and may be either standard functions, or async functions.
    lifespan - A lifespan context function, which can be used to perform startup and shutdown tasks. This is a newer style_path that replaces the on_startup and on_shutdown handlers. Use one or the other, not both.
    """
    def __init__(self, **kwargs):
        self.settings = Settings()
        middleware = kwargs.pop('middleware', [])
        middleware.insert(0, Middleware(SessionMiddleware, secret_key=self.settings.session_secret))
        self.module: str = kwargs.pop('module', 'main')
        self.app_name: str = kwargs.pop('app_name', 'app')
        self.lang: str = kwargs.pop('lang', 'en')
        self.title: str = kwargs.pop('title', 'SpaceStar')
        self.static_directory: Optional[str] = kwargs.pop('static_directory', None)
        self.templates_directory: Optional[str] = kwargs.pop('templates_directory', None)
        self.index_template: Optional[str] = kwargs.pop('index_template', None)
        self._page_content = None
        self._page_title = None
        if self.templates_directory:
            self.templates = Jinja2Templates(
                    directory=os.path.join(os.getcwd(), self.templates_directory),context_processors=[app_context])
        else:
            self.templates = Jinja2Templates(
                    env=jinja2.Environment(), context_processors=[app_context])
            
        if all([self.templates_directory is None, self.index_template is None]):
            raise ValueError('SpaceStar requires either templates_directory or index_template')

        super().__init__(middleware=middleware, **kwargs)
        if self.static_directory:
            self.routes.insert(1, Mount(
                    '/static',
                    app=StaticFiles(directory=os.path.join(os.getcwd(), self.static_directory)),
                    name='static'))
    
    def set_global(self, name, value):
        self.templates.env.globals[name] = value
        
    @property
    def page_content(self):
        try:
            return self._page_content
        finally:
            self._page_content = ''
    
    @page_content.setter
    def page_content(self, value: Markup):
        self._page_content = value
    
    @property
    def page_title(self):
        try:
            return self._page_title
        finally:
            self._page_title = ''
    
    @page_title.setter
    def page_title(self, value: Markup):
        self._page_title = value
        
    
    def from_string(self, source: str):
        return self.templates.env.from_string(source=source, globals=self.globals)
    
    @property
    def globals(self):
        return self.templates.env.globals
    
    @property
    def index_from_string(self):
        return self.from_string(self.index_template)

    @property
    def index(self) -> jinja2.Template:
        if self.templates_directory:
            return self.templates.get_template('index.html')
        return self.index_from_string

    @staticmethod
    def element(tag, *args, **kwargs):
        return Element(tag, *args, **kwargs)

    def render(self, request, / , template: str = None, source: str = None, **kwargs) -> str:
        kwargs['app'] = request.app
        kwargs['request'] = request
        kwargs['id'] = functions.random_id(4)
        if template:
            return self.templates.get_template(template).render(**kwargs)
        elif source:
            return self.from_string(source=source).render(**kwargs)
        return self.index.render(**kwargs)

    def response(self, request: Request, *, template: str = None, source: str = None, **kwargs) -> HTMLResponse:
        return HTMLResponse(self.render(request, template=template, source=source, **kwargs))

    def run(self, *args, **kwargs):
        if args:
            string = ':'.join(args)
        else:
            string = f'{self.module}:{self.app_name}'
        port = kwargs.pop('port', self.settings.port)
        uvicorn.run(string, port=port, **kwargs)
        
    @staticmethod
    def write_query(data: dict) -> str:
        return functions.write_query(data)
    
    @staticmethod
    def markup_detail(instance: SpaceModel):
        return Markup(instance.element_detail())
    
    @staticmethod
    def markup_list_item(instance: SpaceModel):
        return Markup(instance.element_list_group_item())
    
    @staticmethod
    def markup_item_link(instance: SpaceModel, href: str):
        return Markup(instance.element_list_group_item_action(href=href))
    
    @staticmethod
    def markup_htmx_action(instance: SpaceModel, href: str, target: str, indicator: str):
        return Markup(instance.element_list_group_item_htmx_action(href=href, target=target, indicator=indicator))

    def create_route(self, path: str, *, name: str = None, methods: list[str] = None):
        def decorator(endpoint):
            @wraps(endpoint)
            def wrapper():
                self.append(Route(path=path, endpoint=endpoint, name=name, methods=methods or ['GET']))
                return self
            return wrapper()
        return decorator

    def append(self, app: Route | Mount | Router) -> None:
        self.routes.append(app)

    def prepend(self, app: Route | Mount | Router) -> None:
        self.routes.insert(0, app)
        
    def insert(self, index: int, app: Route | Mount) -> None:
        self.routes.insert(index, app)
        
    @staticmethod
    async def process_form_data(request: Request) -> dict:
        form_data: FormData = await request.form()
        data, result = defaultdict(list), {}
        for key, value in form_data.items():
            data[key].append(value)
        for key in data:
            if not key == 'search':
                value = data[key]
                if len(value) == 0:
                    result[key] = None
                elif len(value) == 1:
                    result[key] = value[0]
                else:
                    result[key] = value
        return result
        
    @property
    def E(self):
        return Element


